<p>{{ _fsdomain('Welcome %(email)s!', email=user.email) }}</p>

<p>{{ _fsdomain('You can log into your account through the link below:') }}</p>

<p><a href="{{ login_link }}">{{ _fsdomain('Login now') }}</a></p>
